# Copyright (C) 2019 Hannu Viitala
#
# The source code in this file is released under the MIT license.
# Go to http://opensource.org/licenses/MIT for the full license details.

import upygame as pygame
import glob
import sprite
import urandom as random

# The game object class.
# Used for the enemies.
class GameObject(sprite.Sprite):
    def __init__(self, surfaces, frameOffsets):
        sprite.Sprite.__init__(self)
        self.frames = surfaces
        self.frameOffsets = frameOffsets
        self.currentAnimFrameNum = 0;
        self.image = self.frames[self.currentAnimFrameNum]  # current image
        self.animDur = 3;
        self.animDurCounter = self.animDur;
        self.animDir = 1;

        # velocity and position in the world
        self.vx = 0
        self.vy = 0
        self.x = 0 # world x
        self.y = 0 # world y
        # Note. the rect member data is in viewport domain (e.g. screen)

        # Position in the viewport
        self.rect = self.frames[0].get_rect()

        # Direction to go
        self.dir = 4 # 1=up, 2=right, 3=down, 4=left
        self.randTurnFrameNum = 40
        self.randCheckHeroPosFrameNum = 50

    def setvel(self, vx, vy):
        self.vx = vx
        self.vy = vy

    def update(self):

        # Advance frame if animation is set
        if self.animDur > 0:

            # if animation duration has elapsed, advance frame
            if self.animDurCounter == 0:
                self.currentAnimFrameNum += self.animDir
                if (self.currentAnimFrameNum >= len(self.frames)) or (self.currentAnimFrameNum < 0):
                    self.animDir = - self.animDir
                    self.currentAnimFrameNum += self.animDir

                # animation duration
                self.animDurCounter = self.animDur

                # Set current image
                self.image = self.frames[self.currentAnimFrameNum]
                self.x += self.frameOffsets[self.currentAnimFrameNum][0]
                self.y += self.frameOffsets[self.currentAnimFrameNum][1]
            else:
                self.animDurCounter -= 1

        # Advance position in the world
        prevX = self.x
        prevY = self.y
        if(self.dir == 1): # up
            self.y -= self.vy
        elif(self.dir == 2): # right
            self.x += self.vx
        elif(self.dir == 3): # down
            self.y += self.vy
        else: # right
            self.x -= self.vx

        # If the distance to the hero is bigger that before, change direction
        if( self.randCheckHeroPosFrameNum < glob.frameNum and
            ((abs(glob.heroGob.x-self.x) > abs(glob.heroGob.x-prevX)) or
            (abs(glob.heroGob.y-self.y) > abs(glob.heroGob.y-prevY)))):

            self.x = prevX
            self.y = prevY
            self.dir += 1
            if(self.dir>4): self.dir = 1
            self.randCheckHeroPosFrameNum = random.getrandbits(3) + glob.frameNum + 10

        # From time to time, just turn
        if(self.randTurnFrameNum == glob.frameNum):
            self.dir += 1
            if(self.dir>4): self.dir = 1
            self.randTurnFrameNum = random.getrandbits(4) * 40 + glob.frameNum + 40

        # Check for collisions: to Hero, to the wall or to a bullet
        heroWasHit, enemyWasHit, enemyIsOverWall = self.CheckCollisions2()
        if heroWasHit != None:
            glob.HeroWasHit()

        # if this enemy was hit, set it to out of screen
        if enemyWasHit != None:
            self.x = -100
            self.y = -100

        # if this enemy was hit, set it to out of screen
        if enemyIsOverWall:
            #self.vx = 0
            #self.vy = 0
            self.x = prevX
            self.y = prevY
            self.dir += 1
            if(self.dir>4): self.dir = 1
            #print("isOverWall")

        # Set the position in the viewport
        self.rect.x = self.x + glob.viewPortX
        self.rect.y = self.y + glob.viewPortY

    # Check all collisions
    def CheckCollisions2(self):
        heroWasHit = None
        enemyWasHit = None
        enemyIsOverWall = None
        herocollide = glob.heroGob.rect.colliderect
        heroGobX = glob.heroGob.x
        heroGobY = glob.heroGob.y
        heroGobW = glob.heroGob.rect.width
        heroGobH = glob.heroGob.rect.height
        srect = pygame.Rect(0,0,0,0)

        s = self
        srect.width = s.rect.width
        srect.height = s.rect.height
        srectx = s.x
        srecty = s.y
        srect.x = srectx
        srect.y = srecty

        # Check the collision to the wall in the tilemap
        tileIdList = glob.tilemap.get_tile_ids(srectx, srecty, srectx+srect.width, srecty+srect.height, 8)
        tl = glob.tileAttributeArray[tileIdList[0]] & glob.TILE_FLAG_BLOCKING
        tr = glob.tileAttributeArray[tileIdList[1]] & glob.TILE_FLAG_BLOCKING
        bl = glob.tileAttributeArray[tileIdList[2]] & glob.TILE_FLAG_BLOCKING
        br = glob.tileAttributeArray[tileIdList[3]] & glob.TILE_FLAG_BLOCKING
        if( tl + tr + bl + br > 0):
            enemyIsOverWall = True

        # Check if the enemy is out-of-screen
        if (srectx<0 or srecty<0 ):
            return heroWasHit, enemyWasHit, enemyIsOverWall

        # Check for the hero hit is by the enemy
        if (abs(heroGobX-srectx<heroGobW) and abs(heroGobY-srecty<heroGobH)):
            if herocollide(s.rect):
                heroWasHit = s

        # Check for the enemy hit by the laser
        # Note: this can be optimized by checking only the bullets that are in yhe same quadrant
        # as the enemy, related to player.
        srectw = srect.width
        srecth = srect.height
        for i in range(glob.numLasers):
            laserx = glob.lasers.xlist[i]
            lasery = glob.lasers.ylist[i]
            if (abs(srectx-laserx) < srectw and abs(srecty-lasery) < srecth ):
                # Check with more precision
                laserw = glob.lasers.imagelist[i].get_rect().width
                laserh = glob.lasers.imagelist[i].get_rect().height
                if((srectx < laserx + laserw) and (srecty < lasery + laserh) and
                    (srectx + srectw > laserx) and (srecty + srecth > lasery)):
                    enemyWasHit = s

        return heroWasHit, enemyWasHit, enemyIsOverWall

# The hero object class.
# Used for the enemies.
class HeroGameObject(GameObject):

    def __init__(self, surfaces, frameOffsets):
        GameObject.__init__(self, surfaces, frameOffsets)

    def update(self):

        # Advance frame if animation is set
        if self.animDur > 0:

            # if animation duration has elapsed, advance frame
            if self.animDurCounter == 0:
                self.currentAnimFrameNum += self.animDir
                if (self.currentAnimFrameNum >= len(self.frames)) or (self.currentAnimFrameNum < 0):
                    self.animDir = - self.animDir
                    self.currentAnimFrameNum += self.animDir

                # The animation duration
                self.animDurCounter = self.animDur

                # Set current image
                self.image = self.frames[self.currentAnimFrameNum]
                self.x += self.frameOffsets[self.currentAnimFrameNum][0]
                self.y += self.frameOffsets[self.currentAnimFrameNum][1]
            else:
                self.animDurCounter -= 1

        # Advance position in the world
        prevX = self.x
        prevY = self.y
        self.x += self.vx
        self.y += self.vy

        # Check collisions
        isOverWall = self.CheckCollisions2()

        # If would be over the wall, do not move.
        if isOverWall:
            self.x = prevX
            self.y = prevY
            #print("isOverWall")

        # Set the position in the viewport
        self.rect.x = self.x + glob.viewPortX
        self.rect.y = self.y + glob.viewPortY

    # Check all collisions
    def CheckCollisions2(self):
        isOverWall = None
        srect = pygame.Rect(0,0,0,0)
        s = self
        srect.width = s.rect.width
        srect.height = s.rect.height
        srectx = s.x
        srecty = s.y
        srect.x = srectx
        srect.y = srecty

        # Check the collision to the wall in the tilemap
        tileIdList = glob.tilemap.get_tile_ids(srectx, srecty, srectx+srect.width, srecty+srect.height, 8)
        tl = glob.tileAttributeArray[tileIdList[0]] & glob.TILE_FLAG_BLOCKING
        tr = glob.tileAttributeArray[tileIdList[1]] & glob.TILE_FLAG_BLOCKING
        bl = glob.tileAttributeArray[tileIdList[2]] & glob.TILE_FLAG_BLOCKING
        br = glob.tileAttributeArray[tileIdList[3]] & glob.TILE_FLAG_BLOCKING
        if( tl + tr + bl + br > 0):
            isOverWall = True

        return isOverWall

# The Particle class
# Used for the bullets
class Particle(object):

    def __init__(self, xlist, ylist, vxlist, vylist, imagelist):
        self.partCount = len(xlist)
        self.xlist = xlist
        self.ylist = ylist
        self.vxlist = vxlist
        self.vylist = vylist
        self.imagelist = imagelist

    def update(self):
        # Cache values locally for performance in the loop
        x_list = self.xlist
        y_list = self.ylist
        vx_list = self.vxlist
        vy_list = self.vylist
        for i in range(self.partCount):
            # Advance position
            newX = x_list[i] + vx_list[i]
            newY = y_list[i] + vy_list[i]

            # Check for collision to the wall
            tileId = glob.tilemap.get_tile_id(newX, newY, 8)
            if(glob.tileAttributeArray[ tileId ] & glob.TILE_FLAG_BLOCKING > 0 ):
                # Dismiss the laser
                newX = -100
                newY = -100
                vx_list[i] = 0
                vy_list[i] = 0

            # Set the new values
            x_list[i] = newX
            y_list[i] = newY


    def draw(self, surface):
        # Cache values locally for performance in the loop
        x_list = self.xlist
        y_list = self.ylist
        surface_blit = surface.blit
        image_list = self.imagelist
        part_Count = self.partCount
        view_PortX = glob.viewPortX
        view_PortY = glob.viewPortY
        for i in range(part_Count):
            surface_blit(image_list[i], x_list[i] + view_PortX, y_list[i] + view_PortY)

