# This is a demo of using the Pokitto "Joystick N' Rumble Hat"
# The demos is a twin stick shooter.
# Note: this demo can also be played without the hat.
# Features used
# - The analog joystick in the hat
# - The rumble in the hat
# - A tilemap class

# Copyright (C) 2019 Hannu Viitala
#
# The source code in this file is released under the MIT license.
# Go to http://opensource.org/licenses/MIT for the full license details.

import upygame as pygame
import urandom as random
import glob
import data as spritedata
import GameObjects
import sprite
import gc
import umachine
from micropython import const

SCREEN_W = const(110)
SCREEN_H = const(88)
TILE_W = const(8)
TILE_ITEM_W = const(16)
TILE_H = const(8)
TILEMAP_W = const(16) # in tiles
TILEMAP_H = const(16) # in tiles
TILEMAP_W_ITEMS = const(TILEMAP_W//2) # in items. One item contais 2 tiles.
TILEMAP_W_PX = const(TILEMAP_W*TILE_W)
TILEMAP_H_PX = const(TILEMAP_H*TILE_H)
HERO_VP_MOVE_TLX  = const(40)
HERO_VP_MOVE_TLY  = const(40)
HERO_VP_MOVE_BRX  = const(SCREEN_W-40)
HERO_VP_MOVE_BRY  = const(SCREEN_H-40)

# Joyhat pins
glob.rumbleMotor = umachine.Pin ( umachine.Pin.EXT0, umachine.Pin.OUT )
glob.rumbleMotor.off()
joyhatX = umachine.Pin ( umachine.Pin.EXT1, umachine.Pin.ANALOG_IN )
joyhatY = umachine.Pin ( umachine.Pin.EXT2, umachine.Pin.ANALOG_IN )
#joyhatB1 = umachine.Pin ( umachine.Pin.EXT15, umachine.Pin.IN )
#joyhatB2 = umachine.Pin ( umachine.Pin.EXT14, umachine.Pin.IN )

# Check if this is HW or emulator
test = joyhatY.value()
print("joyhatY.value()=",test)
isHW = True
if(test==0 or test==-1): # emulator do not support pins
    isHW = False

# Check if the hat is used
isJoyhat = False
if(isHW and test>10000): # Supposing the joyhat is not touched when the game starts(!)
    isJoyhat = True
print("isHW=",isHW,", isJoyhat=",isJoyhat)

# Setup screen
pygame.display.init(False)
pygame.display.set_palette_16bit([
    0, 0x001a, 0xd000, 0xd01a, 0x06a0, 0x06ba, 0xd6a0, 0xd6ba,
    0, 0x001f, 0xf800, 0xf81f, 0x07e0, 0x07ff, 0xffe0, 0xffff
]);
screen = pygame.display.set_mode() # full screen

# Setup global variables
glob.screenW = screen.get_rect().width
glob.screenH = screen.get_rect().height
glob.viewPortX = 0
glob.viewPortY = 0

# Set tile attributes
glob.tileAttributeArray = bytearray(16)
glob.tileAttributeArray[3] = glob.TILE_FLAG_BLOCKING
glob.tileAttributeArray[11] = glob.TILE_FLAG_BLOCKING
glob.tileAttributeArray[12] = glob.TILE_FLAG_BLOCKING

# Initialise the tilemap.
glob.tilemap = pygame.tilemap.Tilemap(TILEMAP_W,TILEMAP_H,spritedata.mapPixels)
glob.tilemap.set_tile(0x0, 8, 8, spritedata.blackTileSurf);
glob.tilemap.set_tile(0xb, 8, 8, spritedata.wallHTileSurf);
glob.tilemap.set_tile(0xc, 8, 8, spritedata.wallVTileSurf);
glob.tilemap.set_tile(0x3, 8, 8, spritedata.wallCornerTileSurf);
glob.tilemap.set_tile(0x7, 8, 8, spritedata.shadowHTileSurf);
glob.tilemap.set_tile(0xd, 8, 8, spritedata.shadowVTileSurf);
glob.tilemap.set_tile(0x2, 8, 8, spritedata.shadowCorner1TileSurf);
glob.tilemap.set_tile(0x4, 8, 8, spritedata.shadowCorner2TileSurf);

print('display ready')

# setup sprite group
all_sprites = sprite.Group()
all_enemies = sprite.Group()


# Create hero sprites
glob.heroGob = GameObjects.HeroGameObject(
    [spritedata.pokittoSurf_f1, spritedata.pokittoSurf_f2, spritedata.pokittoSurf_f3],
    [[0,0],[0,0],[0,0]])
glob.heroGob.x = 40
glob.heroGob.y = 30
glob.heroGob.animDur = 2;

# Create enemy sprites
numEnemies = 8
blueenemySurfList = [spritedata.blueenemySurf_f1, spritedata.blueenemySurf_f2, spritedata.blueenemySurf_f3]
enemyOffsetList = [[0,0],[0,0],[0,0]]
redenemySurfList = [spritedata.redenemySurf_f1, spritedata.redenemySurf_f2, spritedata.redenemySurf_f3]
for i in range(numEnemies):

    enemyGob = GameObjects.GameObject(blueenemySurfList, enemyOffsetList)

    # out of screen
    enemyGob.x = -100
    enemyGob.y = -100

    all_sprites.add(enemyGob)
    all_enemies.add(enemyGob)

    enemyGob = GameObjects.GameObject(redenemySurfList,  enemyOffsetList)
    testGob=enemyGob #!!HV

    # out of screen
    enemyGob.x = -100
    enemyGob.y = -100

    all_sprites.add(enemyGob)
    all_enemies.add(enemyGob)

print("enemy count=", len(all_enemies))

# Create laser particles
glob.numLasers = 6
xlist = []
ylist = []
vxlist = []
vylist = []
imagelist  = []
for i in range(glob.numLasers):
    xlist.append(-100)
    ylist.append(-100)
    vxlist.append(0)
    vylist.append(0)
    imagelist.append(spritedata.laserSurf_deg90)
glob.lasers = GameObjects.Particle(xlist, ylist, vxlist, vylist, imagelist)

# init loop variables
vx = 0;
vy = 0;
vp_vy = 0
vp_vx = 0
lastY = 0
shoot = False
lastUsedLaser = 0
lastShotTime = 0
currenFrameTime = 0

# The main loop
while True:

    currenFrameTime = umachine.time_ms()

    # Set the rumble off after the timeout
    if(glob.rumbleEndTime != 0 and currenFrameTime>glob.rumbleEndTime):
        glob.rumbleMotor.off()
        glob.rumbleEndTime = 0

    # Read the Pokitto standard keys
    eventtype = pygame.event.poll()
    if eventtype != pygame.NOEVENT:
        if eventtype.type == pygame.KEYDOWN:
            if eventtype.key == pygame.K_RIGHT:
                vx = 2
            if eventtype.key == pygame.K_LEFT:
                vx = -2
            if eventtype.key == pygame.K_UP:
                vy = -2
            if eventtype.key == pygame.K_DOWN:
                vy = 2
            if eventtype.key == pygame.BUT_A:
                shoot = True

        if eventtype.type == pygame.KEYUP:
            if eventtype.key == pygame.K_RIGHT:
                vx = 0
            if eventtype.key == pygame.K_LEFT:
                vx = 0
            if eventtype.key == pygame.K_UP:
                vy = 0
            if eventtype.key == pygame.K_DOWN:
                vy = 0
            if eventtype.key == pygame.BUT_A:
                shoot = False

    # Store velocity
    vy2 = vy
    vx2 = vx

    if( isHW and isJoyhat):
        # Read Joyhat
        # Value is between 0, 65535
        # Without the hat the value is between 1000 and 3000
        shoot = False
        ly = vy2
        lx = vx2
        joyhat_x = joyhatX.value()
        joyhat_y = joyhatY.value()
    else:
        if(shoot):
            ly = vy2
            lx = vx2
        else:
            ly = 0
            lx = 0
        joyhat_x = 32000
        joyhat_y = 32000

    if(joyhat_x>36000):
        lx = -1
        shoot = True
    elif(joyhat_x<29000):
        lx = 1
        shoot = True
    if(joyhat_y>36000):
        ly = 1
        shoot = True
    elif(joyhat_y<29000):
        ly = -1
        shoot = True

    # Shoot bullets
    if(shoot and (ly!=0 or lx!=0) and (glob.frameNum -lastShotTime) > 1):

        # Get vx and vy for the laser
        vxlist[lastUsedLaser] = 0
        vylist[lastUsedLaser] = 0
        if(ly < 0):
            vylist[lastUsedLaser] = -6
        elif(ly > 0):
            vylist[lastUsedLaser] = 6
        if(lx < 0):
            vxlist[lastUsedLaser] = -6
        elif(lx > 0):
            vxlist[lastUsedLaser] = 6

        # Select the bitmap
        if(lx > 0 and ly==0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg0
        elif(lx > 0 and ly<0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg45
        elif(lx == 0 and ly<0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg90
        elif(lx < 0 and ly<0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg135
        elif(lx < 0 and ly==0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg180
        elif(lx < 0 and ly>0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg225
        elif(lx == 0 and ly>0):
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg270
        else: # (lx > 0 and ly>0)
            imagelist[lastUsedLaser] = spritedata.laserSurf_deg315

        # Get x and y for the laser
        offsetx = 2
        if(glob.frameNum%4 < 2 ):
            offsetx = -2
        offsety = offsetx
        if(lx==ly):
          offsety = - offsetx
        xlist[lastUsedLaser] = glob.heroGob.x + (glob.heroGob.rect.width//2) + offsetx
        ylist[lastUsedLaser] = glob.heroGob.y + (glob.heroGob.rect.height//2) + offsety
        lastUsedLaser += 1
        if(lastUsedLaser>=glob.numLasers):
            lastUsedLaser = 0
        lastShotTime =  glob.frameNum;

    # Set the hero speed
    glob.heroGob.setvel(vx2,vy2);

    # Launch a new bug after N frames
    if (glob.frameNum % 13) == 0:
        # Get first free enemy and set the starting position
        for s in all_enemies:
            if s.x < -16:
                s.x = 110
                y = 10 + random.getrandbits(8) * 60 // 256
                if abs(y - lastY) < 7:
                    if y < lastY:
                        y -= 7
                    else:
                        y += 7
                s.y = y
                lastY = y
                s.setvel(2,2)
                s.currentAnimFrameNum = 0

                break

    # *** UPDATE ***

    # Update hero
    glob.heroGob.update()

    # Calculate viewport position
    hero_vpx = glob.heroGob.rect.x
    dx = HERO_VP_MOVE_TLX - hero_vpx
    vpx = 0
    if(dx < 0):
        vpx = 0  # do not move vp
        dx = HERO_VP_MOVE_BRX - hero_vpx
        if(dx > 0):
            vpx = 0  # do not move vp
        else:
            vpx = dx
    else:
        vpx = dx
    hero_vpy = glob.heroGob.rect.y
    dy = HERO_VP_MOVE_TLY - hero_vpy
    vpy = 0
    if(dy < 0):
        vpy = 0
        dy = HERO_VP_MOVE_BRY - hero_vpy
        if(dy > 0):
            vpy = 0  # do not move vp
        else:
            vpy = dy
    else:
        vpy = dy

    # Move viewport
    newVpX = glob.viewPortX + vpx
    newVpY = glob.viewPortY + vpy

    # Check that viewport is not out bounds
    if(newVpX>0):
        newVpX=0
    elif(newVpX + TILEMAP_W_PX < SCREEN_W):
        newVpX = SCREEN_W - TILEMAP_W_PX
    if(newVpY>0):
        newVpY=0
    elif(newVpY + TILEMAP_H_PX < SCREEN_H):
        newVpY = SCREEN_H - TILEMAP_H_PX
    glob.viewPortX = newVpX
    glob.viewPortY = newVpY

    # update hero's position in the viewport
    glob.heroGob.rect.x = glob.heroGob.x + glob.viewPortX
    glob.heroGob.rect.y = glob.heroGob.y + glob.viewPortY

    # Update enemies
    all_sprites.update()

    # *** DRAW ***

    # Draw tilemap
    glob.tilemap.draw(glob.viewPortX, glob.viewPortY)

    # Draw enemies
    all_sprites.draw(screen)

    # Draw hero
    screen.blit(glob.heroGob.image, glob.heroGob.rect.x, glob.heroGob.rect.y)

    # Update and draw lasers
    glob.lasers.update()
    glob.lasers.draw(screen)

    #text = "jx:"+str(joyhatX.value())
    #umachine.draw_text(0,0,text,15)
    #text = "jy:"+str(joyhatY.value())
    #umachine.draw_text(0,10,text,15)
    #text = "isHW:"+str(isHW)+" isHat:"+str(isJoyhat)
    #umachine.draw_text(0,20,text,15)
    #text = "b1:"+str(joyhatB1.value())+", b2:"+str(joyhatB2.value())
    #umachine.draw_text(0,20,text,1)

    glob.frameNum += 1
    if glob.frameNum > 1000000:
        glob.frameNum = 0;

    pygame.display.flip()


